#include "aligner.hh"
#include "anchor_constraints.hh"
// #include "d_matrix.hh"

#include <math.h>
#include <assert.h>

#include <queue>


using namespace std;

/*
  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  TODO:
  change affine gap-cost computation, one does not need
  a vector for the F matrix, a variable is sufficient
*/


/*
  NOTE on max-diff-am heuristic:
  this heuristic avoids additional effort for the
  computation of arcs with length difference > d
  If the score is computed anyway the arc match
  is still scored
*/


/*
  SEMANTIC OF CONSTRAINTS for local alignment:
  
  currently its possible that constraints are not satisfied
  since their bases don't occur in the local alignment!
  
  => constraints only show their expected behavior in global alignment
*/


/*
  CONSTRAINT HANDLING SOLL NOCH GEÄNDERT WERDEN
  (Dabei beachten: 
  correct handling of constraints in trace-back
  correct handling of constraints for noLP)
  
  constraints can be encoded in scores (currently not done!)

  e.g. 
  * sigma(i,j) = -infty if edge (i,j) is disallowed
  * gapA(i,j) (* means delete i in A after position j in B *)
      = -infty if i is a named position
  
  only encoding constraints in this form can be less efficient
  
  this encoding needs to be supported by the infty arithmetic,
  in particular  a calculation neg_infty+neg_infty-finite must be valid

  advantages:
  we need less if-statements in the main code,
  handling constraints correctly will be easier
*/

/*
recursions for global alignment:

the following recursions simplify the actual locarna algorithm
most notably:
* the computation of matrices M^ab and M^a'b' where al=a'l and bl=bl' is joined
* no constraint handling is done
* no heuristics (max-diff-am)

standard recursion
M^ab_ij = max (M^ab_i-1j-1 + s(i,j);
               M^ab_i-1j + g;
	       M^ab_ij-1 + g;
	       max M^ab_a'l-1b'l-1 + D_a'b': a'b' a'r=i b'r=j)
D_ab    = M^ab_ar-1br-1 + s(a,b)


no lonely pairs
M^ab_ij = max (M^ab_i-1j-1 + s(i,j);
               M^ab_i-1j + g;
	       M^ab_ij-1 + g;
	       max M^ab_a'l-1b'l-1 + D_a'b': a'b' a'r=i b'r=j)
D_ab    = max { M^a'b'_a'r-1b'r-1 + s(a,b) + s(a',b');
                D_a'b' + s(a,b) }
	  where a'l-1=al,b'l-1=bl,a'r+1=ar,b'r+1=br


stacking
M^ab_ij = max ( M^ab_i-1j-1 + s(i,j);
                M^ab_i-1j + g;
	        M^ab_ij-1 + g;
	        max M^ab_a'l-1b'l-1 + D_a'b': a'b' a'r=i b'r=j )
D_ab    = max ( M^ab_ar-1br-1 + s(a,b);
                D^a'b' + s'(a',b') )
	  where a'l-1=al,b'l-1=bl,a'r+1=ar,b'r+1=br
*/


// ------------------------------------------------------------
// Aligner: align / compute similarity
//

Aligner::Aligner(const Aligner &a)
    :scoring(a.scoring),
     params(a.params),
     seqA(a.seqA),
     bpsA(a.bpsA),
     seqB(a.seqB),
     bpsB(a.bpsB),
     r(a.r),
     D(a.D),
     Ms(a.Ms),
     Es(a.Es),
     Fs(a.Fs),
     max_i(a.max_i),
     max_j(a.max_j),
     D_created(a.D_created),
     alignment(a.alignment)
{}

const Aligner &
Aligner::operator = (const Aligner &a) {
    params=a.params;
    r=a.r;
    D=a.D;
    Ms=a.Ms;
    Es=a.Es;
    Fs=a.Fs;
    seqA=a.seqA;
    bpsA=a.bpsA;
    seqB=a.seqB;
    bpsB=a.bpsB;
    alignment=a.alignment;
    max_i=a.max_i;
    max_j=a.max_j;
    D_created=a.D_created;
    scoring=a.scoring;
    
    return a;
}

Aligner::Aligner(Sequence &seqA_, BasePairs &bpsA_,
		 Sequence &seqB_, BasePairs &bpsB_,
		 const AlignerParams *ap_,
		 const Scoring *s_
		 )
    : scoring(s_),params(ap_),
      seqA(seqA_), bpsA(bpsA_), seqB(seqB_), bpsB(bpsB_),
      r(1,1,seqA_.length(),seqB_.length()),
      D_created(false),
      alignment(seqA_,seqB_)
{
    Ms.resize(params->STRUCT_LOCAL?8:1);
    Es.resize(params->STRUCT_LOCAL?4:1);
    Fs.resize(params->STRUCT_LOCAL?4:1);
    
    D.resize(bpsA.num_bps(),bpsB.num_bps());
    D.fill(infty_score_t::neg_infty);
    
    for (size_type k=0; k<(params->STRUCT_LOCAL?8:1); k++) {
	Ms[k].resize(seqA.length()+1,seqB.length()+1);
    }	
    for (size_type k=0; k<(params->STRUCT_LOCAL?4:1); k++) {
	Es[k].resize(seqB.length()+1);
	Fs[k].resize(seqA.length()+1);
    }
}


// standard cases in alignment: base match, base in/del, arc match
// (for structure local alignment this is extended by exclusion handling)
// 
// if lonely basepairs are disallowed, there is special treatment
//
// align_noex has the side effect: computes entry in Es[state] and Fs[state]
infty_score_t
Aligner::align_noex(int state, size_type al, size_type bl, size_type i, size_type j) {
    
    assert(0<=state && state<4);

    ScoreMatrix &M = Ms[state];
    ScoreVector &E = Es[state];
    ScoreVector &F = Fs[state];
    
    // compute E entry
    E[j]=std::max( E[j] + scoring->gapA(i,j),
		   M(i-1,j) + scoring->gapA(i,j) + scoring->indel_opening() );

    // compute F entry
    F[i]=std::max( F[i] + scoring->gapB(i,j),
		   M(i,j-1) + scoring->gapB(i,j) + scoring->indel_opening() );
    
    infty_score_t max_score = infty_score_t::neg_infty;
    
    // base match
    if (params->constraints.allowed_edge(i,j)) {
	max_score = M(i-1,j-1) + scoring->basematch(i,j);
    }
    
    // base del
    if (! params->constraints.aligned_in_a(i)) {
	max_score=max(max_score, E[j]);
    }
    
    // base ins
    if (! params->constraints.aligned_in_b(j)) {
	max_score=max(max_score, F[i]);
    }
    
    // arc match
    
    // standard case for arc match (without restriction to lonely pairs)
    //
    
    if (params->constraints.allowed_edge(i,j))
    {
    	const BasePairs::RightAdjList &adjlA = bpsA.right_adjlist(i);
	const BasePairs::RightAdjList &adjlB = bpsB.right_adjlist(j);
	
	// for all pairs of arcs in A and B that have right ends i and j, respectively
	//
	for (BasePairs::RightAdjList::const_iterator arcA=adjlA.begin();
	     arcA!=adjlA.end() && arcA->left() > al; ++arcA) {
	    for (BasePairs::RightAdjList::const_iterator arcB=adjlB.begin();
		 arcB!=adjlB.end() && arcB->left() > bl ; ++arcB) {
		//
		// consider score for match of basepairs
		
		if (params->constraints.allowed_edge(arcA->left(),arcB->left())) {
		    max_score =
			max(max_score,
			    M(arcA->left()-1,arcB->left()-1)
			    + D(arcA->idx(),arcB->idx())
			    );
		}
	    }
	}
    }
    
    return max_score.normalized_neg();
}


/*
  BUG: init has to be corrected for structure locality, see Technical Details
*/

// generic initalization method.
// Depending on the state
// the 0th row and column are initialized
//
void
Aligner::init_state(int state, int al, int ar, int bl, int br, 
		    bool globalA, bool exclA,
		    bool globalB, bool exclB) {
    
    ScoreMatrix &M = Ms[state];
    
    M(al,bl)=(infty_score_t)0;
    
    if (state<4) {
	ScoreVector &E = Es[state];
	ScoreVector &F = Fs[state];
	
	for (int i=al; i<ar; i++) {
	    F[i]=infty_score_t::neg_infty;
	}
	for (int j=bl; j<br; j++) {
	    E[j]=infty_score_t::neg_infty;
	}
    }
    
    score_t indel_score=scoring->indel_opening();
    for (int i=al+1; i<ar; i++) {
	indel_score += scoring->gapA(i,0);
	M(i,bl) =
	    (infty_score_t)(exclA
			    ? infty_score_t::neg_infty
			    : (globalA
			       ? (infty_score_t)indel_score
			       : (infty_score_t)0));
    }
    
    indel_score=scoring->indel_opening();
    for (int j=bl+1; j<br; j++) {
	indel_score += scoring->gapB(0,j);
	M(al,j) =
	    (infty_score_t)(exclB
			    ? infty_score_t::neg_infty
			    : (globalB
			       ? (infty_score_t)indel_score
			       : (infty_score_t)0));
    }
}

void
Aligner::align_in_arcmatch(int al,int ar,int bl,int br,
				bool allow_exclusion) {

    // cout << al << " " << ar <<" " << bl << " " << br <<endl;

    // if in a sequence the state is not open than gaps with cost scoring->gap() have to be introduced.
    // In open cases the maximal similarity is 0 in the row/column,
    // since an exclusion can be introduced.

    init_state(E_NO_NO,al,ar,bl,br,true ,false,true ,false);
    
    if (allow_exclusion) {
	init_state(E_X_NO, al,ar,bl,br, true , true , true , false);
	init_state(E_NO_X, al,ar,bl,br, true , false, true , true );
	init_state(E_X_X,  al,ar,bl,br, true , true , true , true );
	init_state(E_OP_NO,al,ar,bl,br, true , true , true , false);
	init_state(E_NO_OP,al,ar,bl,br, true , false, true , true );
	init_state(E_X_OP, al,ar,bl,br, true , true , true , true );
	init_state(E_OP_X, al,ar,bl,br, true , true , true , true );
    }

    FOR_IJ_IN_ARCMATCH
	Ms[E_NO_NO](i,j)=align_noex(E_NO_NO,al,bl,i,j);
    
    if (allow_exclusion) {
	int state;

	state=E_OP_NO;
	FOR_IJ_IN_ARCMATCH { Ms[state](i,j) = max(Ms[state](i-1,j), Ms[E_NO_NO](i,j)); }
	    
	state=E_NO_OP;
	FOR_IJ_IN_ARCMATCH { Ms[state](i,j) = max(Ms[state](i,j-1), Ms[E_NO_NO](i,j)); }

	state=E_NO_X;
	FOR_IJ_IN_ARCMATCH { Ms[state](i,j) = max(align_noex(state,al,bl,i,j),
						  Ms[E_NO_OP](i,j)+scoring->exclusion()); }
	state=E_OP_X;
	FOR_IJ_IN_ARCMATCH { Ms[state](i,j) = max(Ms[state](i-1,j), Ms[E_NO_X](i,j)); }
	    
	state=E_X_NO;
	FOR_IJ_IN_ARCMATCH { Ms[state](i,j) = max(align_noex(state,al,bl,i,j),
						  Ms[E_OP_NO](i,j)+scoring->exclusion()); }
	
	state=E_X_OP;
	FOR_IJ_IN_ARCMATCH { Ms[state](i,j) = max(Ms[state](i,j-1), Ms[E_X_NO](i,j)); }
	    
	state=E_X_X;
	FOR_IJ_IN_ARCMATCH { Ms[state](i,j) = max(align_noex(state,al,bl,i,j),
						  max(Ms[E_OP_X](i,j)+scoring->exclusion(),
						      Ms[E_X_OP](i,j)+scoring->exclusion())); }
    }
}

// compute the entries in the D matrix that
// can be computed from the matrix/matrices M
// for the subproblem al,bl,max_ar,max_br
//
// pre: M matrices are computed by a call to 
//      align_in_arcmatch(al,max_ar,bl,max_br,params->STRUCT_LOCAL)
//
void 
Aligner::fill_D_entries(size_type al, size_type bl,
			const BasePairs::LeftAdjList &adjlA,
			const BasePairs::LeftAdjList &adjlB,
			size_type max_ar, size_type max_br)
{
    for (BasePairs::LeftAdjList::const_iterator arcA=adjlA.begin();
	 arcA!=adjlA.end(); ++arcA) {
	for (BasePairs::LeftAdjList::const_iterator arcB=adjlB.begin();
	     arcB!=adjlB.end(); ++arcB) {
	    size_type ar = arcA->right();
	    size_type br = arcB->right();
	    //
	    // if right ends ar,br exceed the limits max_ar,max_br resp.
	    // the corresponding D entry is set to neg_infty
	    // in order to dissalow the arc match completely.
	    // This occurs only due to the max-diff-am heuristic.
	    if (ar>max_ar || br>max_br ) {
		D(arcA->idx(),arcB->idx()) = infty_score_t::neg_infty;
	    } else {
		infty_score_t m=Ms[0](ar-1,br-1);
		if (params->STRUCT_LOCAL) {
		    // if we align structure local
		    // we need to determine the maximum of entries in Ms[k],
		    // where k is a non-open state
		    for (size_type k=1; k<4; k++) {
			m=max(m, Ms[k](ar-1,br-1));
		    }
		}
		
		D(arcA->idx(),arcB->idx()) = m + scoring->arcmatch(*arcA,*arcB);
		
		if (scoring->stacking()) {
		    if (bpsA.exists_arc(al+1,ar-1) && bpsB.exists_arc(bl+1,br-1)) { 
			const Arc &arcAI = bpsA.arc(al+1,ar-1); // inner arc in a
			const Arc &arcBI = bpsB.arc(bl+1,br-1); // inner arc in b
			
			//std::cout <<"stacking "<< (*arcA) << " " << (*arcB) << " " << D(arcA->idx(),arcB->idx()) << " " << D(arcAI.idx(),arcBI.idx()) + scoring->arcmatch_stacked(*arcA,*arcB) << " " << scoring->arcmatch(*arcA,*arcB) << " " << scoring->arcmatch_stacked(*arcA,*arcB) << std::endl;

			D(arcA->idx(),arcB->idx()) =
			    max(D(arcA->idx(),arcB->idx()),
				D(arcAI.idx(),arcBI.idx()) + scoring->arcmatch_stacked(*arcA,*arcB)
			    );
		    }
		}
	    }
	}
    }
}

void 
Aligner::fill_D_entries_noLP(size_type al, size_type bl, size_type max_ar, size_type max_br) {
    // get adj lists of arcs starting in al-1, bl-1
    
    const BasePairs::LeftAdjList &adjlA = bpsA.left_adjlist(al-1);
    const BasePairs::LeftAdjList &adjlB = bpsB.left_adjlist(bl-1);
    
    for (BasePairs::LeftAdjList::const_iterator arcA=adjlA.begin();
	 arcA!=adjlA.end(); ++arcA) {
	for (BasePairs::LeftAdjList::const_iterator arcB=adjlB.begin();
	     arcB!=adjlB.end(); ++arcB) {
	    
	    size_type ar = arcA->right()-1;
	    size_type br = arcB->right()-1;

	    //
	    // if right ends ar,br exceed the limits max_ar,max_br resp.
	    // the corresponding D entry is set to neg_infty
	    // in order to dissalow the arc match completely.
	    // This occurs only due to the max-diff-am heuristic.
	    if (ar>max_ar || br>max_br ) {
		D(arcA->idx(),arcB->idx()) = infty_score_t::neg_infty;
	    } else {
		if (bpsA.exists_arc(al,ar) && bpsB.exists_arc(bl,br)) { 
		    const Arc &arcAI = bpsA.arc(al,ar); // inner arc in a
		    const Arc &arcBI = bpsB.arc(bl,br); // inner arc in b
		    
		    infty_score_t m=Ms[0](ar-1,br-1);
		    if (params->STRUCT_LOCAL) {
			// if we align structure local
			// we need to determine the maximum of entries in Ms[k],
			// where k is a non-open state
			for (size_type k=1; k<4; k++) {
			    m=max(m, Ms[k](ar-1,br-1));
			}
		    }
		    
		    D(arcA->idx(),arcB->idx()) = 
			max(m + scoring->arcmatch(arcAI,arcBI),
			    D(arcAI.idx(),arcBI.idx()))
			+ scoring->arcmatch(*arcA,*arcB,scoring->stacking());
		}
	    }
	}
    }
}



// compute all entries D
void
Aligner::align_D() {
    // ------------------------------------------------------------
    // General workflow:
    //
    // for all combinations of left arc ends al and ar
    // 1.) determine for which arc-pairs the D entries can be computed   
    // in one run, 2.) call align_in_arcmatch 3.) call fill_D_entries
    // ------------------------------------------------------------

    // ------------------------------------------------------------
    // traverse the left ends al,bl of arcs in descending order
    //
    for (size_type al=r.get_endA(); al+1>r.get_startA(); al--) {
	const BasePairs::LeftAdjList &adjlA = bpsA.left_adjlist(al);
	if (adjlA.size()>0) {
	    for (size_type bl=r.get_endB(); bl+1>r.get_startB(); bl--) {
		const BasePairs::LeftAdjList &adjlB = bpsB.left_adjlist(bl);
		if (adjlB.size()>0) {
		    // ------------------------------------------------------------
		    // get maximal right ends of arcs with left ends al,bl 
		    // where max_diff_am conditions hold
		    // and no_lonely_pairs condition holds
		    //
		    size_type max_ar=al;
		    size_type max_br=bl;
		    if (params->max_diff_am<0 && !params->no_lonely_pairs) {
		        max_ar=adjlA[adjlA.size()-1].right();
		        max_br=adjlB[adjlB.size()-1].right();
		    } else {
			for (BasePairs::LeftAdjList::const_iterator arcA = adjlA.begin();
			     arcA!=adjlA.end(); ++arcA) {
			    for (BasePairs::LeftAdjList::const_iterator
				     arcB = adjlB.begin();
				 arcB!=adjlB.end(); ++arcB) {
				size_type ar = arcA->right();
				size_type br = arcB->right();
				
				if ((params->max_diff_am<0 
				     || std::max(ar-al,br-bl)-std::min(ar-al,br-bl) <= (size_type)params->max_diff_am)
				    &&
				    (!params->no_lonely_pairs
				     || (bpsA.exists_arc(al-1,ar+1) // there is an enclosing basepair match
					 && bpsB.exists_arc(bl-1,br+1))) 
				    ) {
  				    max_ar=max(max_ar,ar);
				    max_br=max(max_br,br);
				}
			    }
			}
		    }
		    
		    // ------------------------------------------------------------
		    // align under the maximal pair of arcs
		    //
		    align_in_arcmatch(al,max_ar,bl,max_br,params->STRUCT_LOCAL);
		    
		    // ------------------------------------------------------------
		    // fill D matrix entries
		    //
		    if (params->no_lonely_pairs) {
			fill_D_entries_noLP(al,bl,max_ar,max_br);
		    } else {
			fill_D_entries(al,bl,adjlA,adjlB,max_ar,max_br);
		    }
		}
	    }
	}
    }
    
    D_created=true; // now the matrix D is built up
}


// align the top level in case of sequence local alignment
//
infty_score_t
Aligner::align_top_level_locally() {
    //std::cout << r << std::endl;
    
    ScoreMatrix &M=Ms[E_NO_NO];
    infty_score_t max_score=(infty_score_t)0;
    init_state(E_NO_NO,r.get_startA()-1,r.get_endA()+1,r.get_startB()-1,r.get_endB()+1,false,false,false,false);
    
    for (size_type i=r.get_startA(); i<=r.get_endA(); i++) {
	for (size_type j=r.get_startB(); j<=r.get_endB(); j++) {
	    
	    M(i,j) = 
		max( (infty_score_t)0,
		     align_noex( E_NO_NO,r.get_startA()-1,r.get_startB()-1,i,j )
		     );
	    
	    if (max_score < M(i,j)) {
		max_score=M(i,j);
		max_i = i;
		max_j = j;
	    }
	}
    }
    
    
    return max_score;
}


// special top level alignment for the scanning version
infty_score_t
Aligner::align_top_level_localB() {
    // std::cout <<"align local B " << r.get_startA() << " " << r.get_startB() << " "
    //           << r.get_endA() << " " << r.get_endB() << std::endl;
    
    ScoreMatrix &M=Ms[E_NO_NO];
    infty_score_t max_score=infty_score_t::neg_infty;
    init_state(E_NO_NO,r.get_startA()-1,r.get_endA()+1,r.get_startB()-1,r.get_endB()+1,true,false,false,false);
    
    for (size_type i=r.get_startA(); i<=r.get_endA(); i++) {
	for (size_type j=r.get_startB(); j<=r.get_endB(); j++) {
	    M(i,j) = 
		max( (infty_score_t)0,
		     align_noex( E_NO_NO, 
				 r.get_startA()-1,r.get_startB()-1,i,j ) );
	    if (i==r.get_endA() && max_score < M(i,j)) {
		max_score=M(i,j);
		max_i = i;
		max_j = j;
	    }
	}
    }
    return max_score;
}


// compute the alignment score
infty_score_t
Aligner::align() {
    // ------------------------------------------------------------
    // computes D matrix and then
    // does the alignment on the top level
    // ------------------------------------------------------------
    
    if (!D_created) align_D();

    if (params->SEQU_LOCAL) {
	return align_top_level_locally();
    } else { // sequence global alignment
	//
	// align the subsequences that are specified by the restriction object
	//
	align_in_arcmatch(r.get_startA()-1,r.get_endA()+1,r.get_startB()-1,r.get_endB()+1,false);
	max_i=r.get_endA();
	max_j=r.get_endB();
	//
	return Ms[E_NO_NO](max_i,max_j);
    }
}

// ------------------------------------------------------------
// Aligner: traceback

void Aligner::trace_arcmatch(const Arc &arcA, const Arc &arcB) {
    
    size_type al=arcA.left();
    size_type ar=arcA.right();
    size_type bl=arcB.left();
    size_type br=arcB.right();
    
    // --------------------
    // handle case of stacking
    if ( scoring->stacking() ) {
	if (bpsA.exists_arc(al+1,ar-1) 
	    && bpsB.exists_arc(bl+1,br-1)) { 
	    const Arc &arcAI = bpsA.arc(al+1,ar-1); // inner arc in a
	    const Arc &arcBI = bpsB.arc(bl+1,br-1); // inner arc in b
	    if (D(arcA.idx(),arcB.idx()) == D(arcAI.idx(),arcBI.idx()) + scoring->arcmatch_stacked(arcA,arcB)) {
		
		alignment.add_basepairA(arcAI.left(),arcAI.right());
		alignment.add_basepairB(arcBI.left(),arcBI.right());
		alignment.append(arcAI.left(),arcBI.left());
		
		trace_arcmatch(arcAI,arcBI);
		
		alignment.append(arcAI.right(),arcBI.right());
		
		return;
	    }
	}
    }
    
    // --------------------
    // now handle the case that arc match is not stacked
    
    // first recompute M
    align_in_arcmatch(al,ar, bl,br,
		      params->STRUCT_LOCAL);
    
    // then, trace in new M
    if (!params->STRUCT_LOCAL) {
	trace_in_arcmatch(0,al,ar-1,bl,br-1,false);
    } else {
	for (size_type k=0; k<4; k++) {
	    if (D(arcA.idx(),arcB.idx()) == Ms[k](ar-1,br-1) + scoring->arcmatch(arcA,arcB)) {
		trace_in_arcmatch(k,al,ar-1,bl,br-1,false);
		break;
	    }
	}
    }
    return;
}

void Aligner::trace_arcmatch_noLP(const Arc &arcA, const Arc &arcB) {
    assert(bpsA.exists_arc(arcA.left()+1,arcA.right()-1));
    assert(bpsB.exists_arc(arcB.left()+1,arcB.right()-1));

    const Arc &arcAI = bpsA.arc(arcA.left()+1,arcA.right()-1); // inner arc in a
    const Arc &arcBI = bpsB.arc(arcB.left()+1,arcB.right()-1); // inner arc in b
    
    alignment.add_basepairA(arcAI.left(),arcAI.right());
    alignment.add_basepairB(arcBI.left(),arcBI.right());
    alignment.append(arcAI.left(),arcBI.left());
    
    if (D(arcA.idx(),arcB.idx()) == D(arcAI.idx(),arcBI.idx()) + scoring->arcmatch(arcA,arcB,scoring->stacking())) {
	trace_arcmatch_noLP(arcAI,arcBI);
    } else {
	// first recompute M
	align_in_arcmatch(arcAI.left(),arcAI.right(), arcBI.left(),arcBI.right(),
			  params->STRUCT_LOCAL);
	
	// then, trace in new M
	if (!params->STRUCT_LOCAL) {
	    trace_in_arcmatch(0,arcAI.left(),arcAI.right()-1,arcBI.left(),arcBI.right()-1,false);
	} else {
	    for (size_type k=0; k<4; k++) {
		if (D(arcA.idx(),arcB.idx()) == 
		    Ms[k](arcAI.right()-1,arcBI.right()-1)
		    + scoring->arcmatch(arcA,arcB,scoring->stacking()) + scoring->arcmatch(arcAI,arcBI))
		{
		    trace_in_arcmatch(k,arcAI.left(),arcAI.right()-1,arcBI.left(),arcBI.right()-1,false);
		    break;
		}
	    }
	}
    }
    alignment.append(arcAI.right(),arcBI.right());
}

// trace and handle all cases that do not involve exclusions
void
Aligner::trace_noex(int state,size_type oal,size_type i,size_type obl,size_type j,bool tl) {
    ScoreMatrix &M=Ms[state];
    
    // determine where we get M(i,j) from
    
    // match
    if ( params->constraints.allowed_edge(i,j)
	 && M(i,j) ==  M(i-1,j-1)+scoring->basematch(i,j) ) {
	trace_in_arcmatch(state,oal,i-1,obl,j-1,tl);
	alignment.append(i,j);
	return;
    }

    if ( scoring->indel_opening() == 0 ) { // base del and ins, linear cost
	// del
	if ( (!params->constraints.aligned_in_a(i))
	     && M(i,j) == M(i-1,j)+scoring->gapA(i,j)) {
	    trace_in_arcmatch(state,oal,i-1,obl,j,tl);
	    alignment.append(i,-1);
	    return;
	}
	// ins
	if ( (!params->constraints.aligned_in_b(j))
	     && M(i,j) == M(i,j-1)+scoring->gapB(i,j)) {
	    trace_in_arcmatch(state,oal,i,obl,j-1,tl);
	    alignment.append(-1,j);
	    return;
	}
    } else { // base del and ins, affine cost
	// since, we didn't store all values in the E and F matrix
	// we do the traceback in linear time per entry
	// base del
	score_t gap_cost=scoring->indel_opening();
	for (size_type k=1;
	     (i >= oal+k)
		 && (! params->constraints.aligned_in_a(i-k+1));
	     k++)
	{
	    gap_cost += scoring->gapA(i-k+1,j);
	    
	    if ( M(i,j) == M(i-k,j) + gap_cost) {
		// gap in A of length k
		trace_in_arcmatch(state,oal,i-k,obl,j,tl);
		for (size_type l=k;l>0;l--) {
		    alignment.append(i-l+1,-1);
		}
		return;
	    }
	}
	
	// base ins
	gap_cost=scoring->indel_opening();
	for (size_type k=1; 
	     (j >= obl+k)
		 && (! params->constraints.aligned_in_b(j-k+1));
	     k++)
	{
	    gap_cost += scoring->gapB(i,j-k+1);
	    if (M(i,j) == M(i,j-k) + gap_cost) {
		// gap in B of length k
		trace_in_arcmatch(state,oal,i,obl,j-k,tl);
		for (size_type l=k;l>0;l--) {
		    alignment.append(-1,j-l+1);
		}
		
		return;
	    }
	}
    }

    // only consider arc match cases if edge (i,j) is allowed!
    if ( ! params->constraints.allowed_edge(i,j) ) return;
    
    // here (i,j) is allowed
    
    //  arc match
    const size_type &ar=i;
    const size_type &br=j;
    
    const BasePairs::RightAdjList &adjlA = bpsA.right_adjlist(ar);
    const BasePairs::RightAdjList &adjlB = bpsB.right_adjlist(br);
    
    for (BasePairs::RightAdjList::const_iterator arcA=adjlA.begin();
	 arcA!=adjlA.end() && arcA->left() > oal; ++arcA) {
	//
	for (BasePairs::RightAdjList::const_iterator arcB=adjlB.begin();
	     arcB!=adjlB.end() && arcB->left() > obl ; ++arcB) {
	    //
	    const size_type &al=arcA->left();
	    const size_type &bl=arcB->left();
	    
	    // if edge (between left ends) is disallowed, then skip arc pair
	    if (! params->constraints.allowed_edge(al,bl)) continue;
	    
	    if ( M(i,j) == 
		 M(al-1,bl-1)
		 + D(arcA->idx(),arcB->idx())) {
		//
		// do the trac for alignment left of the arc match
		trace_in_arcmatch(state,oal,al-1,obl,bl-1,tl);
		
		//cout << "arcmatch "<<(al)<<","<<i<<";"<<(bl)<<","<<j<<" :: "
		//      <<(arcA->w)<<" + "<<(arcB->w)<< " + " << tau(al,bl,i,j)  <<endl;
		
		alignment.add_basepairA(al,ar);
		alignment.add_basepairB(bl,br);
		alignment.append(al,bl);
		
		// do the trace below the arc match
		
		if (params->no_lonely_pairs) {
		    trace_arcmatch_noLP(*arcA,*arcB);
		} else {
		    trace_arcmatch(*arcA,*arcB);
		}

		alignment.append(ar,br);
		
		return;
	    }
	}
    }
}

// do the trace within one arc match.
// the cases without exclusions are delegated to trace_noex
void
Aligner::trace_in_arcmatch(int state,int al,int i,int bl,int j,bool tl) {
    //pre: M matrices for arc computed
    ScoreMatrix &M=Ms[state];

    /*string state_text[]={"E_NO_NO", "E_X_NO", "E_NO_X", "E_X_X",
      "E_OP_NO", "E_NO_OP", "E_OP_X", "E_X_OP"};
      cout << "trace_in_arcmatch "<<state_text[state]<<" al:"<<al<<" i:"<<i
      <<" bl:"<<bl<<" j:"<<j<<" :: "<< M(i,j) <<endl;
    */
    
    // terminate traceback if
    // * trace on toplevel (tl)
    // * entry 0
    if (tl && M(i,j)==(infty_score_t)0) {
	return;
    }
    if (i<=al) {
	if (state==E_NO_NO
	    || state==E_OP_NO
	    || state==E_X_NO) {
	    for (int k=bl+1;k<=j;k++)
		alignment.append(-1,k);
	} else {
	    //cout << "exclude B "<<state<<" "<< bl<<" - "<<j<<endl;
	}
	return;
    }

    if (j<=bl) {
	if (state==E_NO_NO
	    || state==E_NO_OP
	    || state==E_NO_X) {
	    for (int k=al+1;k<=i;k++) {
		alignment.append(k,-1);
	    }
	} else {
	    //cout << "exclude A "<<state<<" "<< al<<" - "<<i<<endl;
	}
	return;
    }


    switch(state) {
    case E_NO_NO:
	trace_noex(state,al,i,bl,j,tl);
	break;
    case E_OP_NO:
	if (M(i,j) == M(i-1,j)) {
	    // cout << "exclude A "<<i<<endl;
	    trace_in_arcmatch(state,al,i-1,bl,j,tl);
	} else if (M(i,j) == Ms[E_NO_NO](i,j))
	    trace_in_arcmatch(E_NO_NO,al,i,bl,j,tl);
	break;
    case E_NO_OP:
	if (M(i,j) == M(i,j-1)) {
	    // cout << "exclude B "<<j<<endl;
	    trace_in_arcmatch(state,al,i,bl,j-1,tl);
	} else if (M(i,j) == Ms[E_NO_NO](i,j))
	    trace_in_arcmatch(E_NO_NO,al,i,bl,j,tl);
	break;
    case E_NO_X:
	if (M(i,j) == Ms[E_NO_OP](i,j)+scoring->exclusion()) {
	    trace_in_arcmatch(E_NO_OP,al,i,bl,j,tl);
	} else
	    trace_noex(state,al,i,bl,j,tl);
	break;
    case E_OP_X:
	if (M(i,j) == M(i-1,j)) {
	    //cout << "exclude A "<<i<<endl;
	    trace_in_arcmatch(state,al,i-1,bl,j,tl);
	} else if (M(i,j) == Ms[E_NO_X](i,j))
	    trace_in_arcmatch(E_NO_X,al,i,bl,j,tl);
	break;
    case E_X_NO:
	if (M(i,j) == Ms[E_OP_NO](i,j)+scoring->exclusion()) {
	    trace_in_arcmatch(E_OP_NO,al,i,bl,j,tl);
	} else
	    trace_noex(state,al,i,bl,j,tl);
	break;
    case E_X_OP:
	if (M(i,j) == M(i,j-1)) {
	    // cout << "exclude B "<<j<<endl;
	    trace_in_arcmatch(state,al,i,bl,j-1,tl);
	} else if (M(i,j) == Ms[E_X_NO](i,j))
	    trace_in_arcmatch(E_X_NO,al,i,bl,j,tl);
	break;
    case E_X_X:
	if (M(i,j) == Ms[E_OP_X](i,j)+scoring->exclusion()) {
	    trace_in_arcmatch(E_OP_X,al,i,bl,j,tl); 
	} else if (M(i,j) == Ms[E_X_OP](i,j)+scoring->exclusion()) {
	    trace_in_arcmatch(E_X_OP,al,i,bl,j,tl);
	} else
	    trace_noex(state,al,i,bl,j,tl);
	break;
    }
}


void
Aligner::trace() {
    // pre: last call align_in_arcmatch(r.get_startA()-1,r.get_endA()+1,r.get_startB()-1,r.get_endB()+1);
    //      or align_top_level_locally for SEQU_LOCAL alignment
    
    // reset the alignment strings (to empty strings)
    // such that they can be written again during the trace
    alignment.clear();
    
    trace_in_arcmatch(E_NO_NO,r.get_startA()-1,max_i,r.get_startB()-1,max_j,params->SEQU_LOCAL);
}


/* ------------------------------------------------------------
   Compute k-best alignments by intervall splitting
*/

template <class T>
class greater_second {
public:
    bool operator() (const T& a, const T&b) {
	return a.second < b.second;
    }
};

typedef std::pair<AlignerRestriction,infty_score_t> task_t;

//! write the k best local alignments
void Aligner::kbest(size_type k, 
		    size_type output_width, 
		    bool opt_verbose,
		    bool opt_local_output,
		    bool opt_pos_output
		    ) {
  Aligner &a=*this;

    // compute alignment score for a
    infty_score_t a_score = a.align();
    
    // make priority queue tasks
    std::priority_queue<task_t, vector<task_t>, greater_second<task_t> > tasks;
    
    // put a into tasks
    tasks.push(task_t(a.get_restriction(),a_score));

    for (size_type i=1; i<=k; i++) {
	//std::cout << "queue size: "<<tasks.size()<<std::endl;
	
	// get best task from tasks
	task_t task = tasks.top();

	// pop topmost element
	tasks.pop();
	
	AlignerRestriction &task_r=task.first;
	infty_score_t task_score=task.second;
	
	a.set_restriction(task_r);
	a.align(); // alignment needs to be recomputed (this is fast)!
	// do backtrace and write tasks alignment
	a.trace();
	Alignment alignment = a.get_alignment();
	
	// std::cout << "SCORE: " << task_score << std::endl;

	alignment.write(cout, output_width, task_score, 
			opt_local_output, opt_pos_output);
	
	if (!opt_pos_output) std::cout << std::endl
				       << std::endl;

	if (i==k) break; // break if enough solutions generated
	
	// split the longer sequence according to local alignment
	size_type lenA=task_r.get_endA()-task_r.get_startA();
	size_type lenB=task_r.get_endB()-task_r.get_startB();
	
	// make two clones of AlignerRestriction task_r
	AlignerRestriction r1(task_r);
	AlignerRestriction r2(task_r);
	
	if (lenA>lenB) {
	    // split A
	    int splitA = (alignment.get_local_startA() + alignment.get_local_endA())/2;
	    if (opt_verbose) std::cout <<"Split A at "<<splitA<<std::endl;
	    r1.set_endA(splitA);
	    r2.set_startA(splitA);
	} else {
	    int splitB = (alignment.get_local_startB() + alignment.get_local_endB())/2;
	    if (opt_verbose) std::cout <<"Split B at "<<splitB<<std::endl;
	    r1.set_endB(splitB);
	    r2.set_startB(splitB);
	}
	
	// compute alignment scores for both splits
	
	a.set_restriction(r1);
	infty_score_t a1_score = a.align();
	a.set_restriction(r2);
	infty_score_t a2_score = a.align();
	
	// std::cout <<"a1_score: " << a1_score << std::endl;
	// std::cout <<"a2_score: " << a2_score << std::endl;
	
	
	// put both splits into <tasks>
	
	tasks.push(task_t(r1,a1_score));
	tasks.push(task_t(r2,a2_score));
	
    }
}


// ------------------------------------------------------------
