

#include "aligner_p.hh"
#include "anchor_constraints.hh"
// #include "d_matrix.hh"

#include <math.h>
#include <assert.h>

#include <queue>

using namespace std;

/*
 NOTE on max-diff-am heuristic:
 this heuristic avoids additional effort for the
 computation of arcs with length difference > d
 If the score is computed anyway the arc match
 is still scored
 */

/*
 SEMANTIC OF CONSTRAINTS for local alignment:

 currently its possible that constraints are not satisfied
 since their bases don't occur in the local alignment!

 => constraints only show their expected behavior in global alignment
 */

/*
 CONSTRAINT HANDLING SOLL NOCH GEÄNDERT WERDEN
 (Dabei beachten: 
 correct handling of constraints in trace-back
 correct handling of constraints for noLP)

 constraints can be encoded in scores (currently not done!)

 e.g. 
 * sigma(i,j) = -infty if edge (i,j) is disallowed
 * gapA(i,j) (* means delete i in A after position j in B *)
 = -infty if i is a named position

 only encoding constraints in this form can be less efficient

 this encoding needs to be supported by the infty arithmetic,
 in particular  a calculation neg_infty+neg_infty-finite must be valid

 advantages:
 we need less if-statements in the main code,
 handling constraints correctly will be easier
 */

/*
 recursions for global alignment:

 the following recursions simplify the actual locarna algorithm
 most notably:
 * the computation of matrices M^ab and M^a'b' where al=a'l and bl=bl' is joined
 * no constraint handling is done
 * no heuristics (max-diff-am)

 standard recursion
 M^ab_ij = max (M^ab_i-1j-1 + s(i,j);
 M^ab_i-1j + g;
 M^ab_ij-1 + g;
 max M^ab_a'l-1b'l-1 + D_a'b': a'b' a'r=i b'r=j)
 D_ab    = M^ab_ar-1br-1 + s(a,b)

 */

// ------------------------------------------------------------
// AlignerP: align / compute similarity
//
//=================================================================================================================
//	alignerP()
//=================================================================================================================	
AlignerP::AlignerP(const AlignerP &a) :
	scoring(a.scoring), params(a.params), seqA(a.seqA), bpsA(a.bpsA),
	seqB(a.seqB), bpsB(a.bpsB), r(a.r), D(a.D), M(a.M),
	MPrime(a.MPrime), DPrime(a.DPrime), Mrev(a.Mrev),
	max_i(a.max_i), max_j(a.max_j), D_created(a.D_created),
	DPrime_created(a.DPrime_created),
	alignment(a.alignment) {
}

//=================================================================================================================
//	alignerP()
//=================================================================================================================	
const AlignerP & AlignerP::operator = (const AlignerP &a) {
	params=a.params;
	r=a.r;
	D=a.D;
	M=a.M;
	DPrime=a.DPrime;
	MPrime=a.MPrime;
	Mrev=a.Mrev;
	seqA=a.seqA;
	bpsA=a.bpsA;
	seqB=a.seqB;
	bpsB=a.bpsB;
	alignment=a.alignment;
	max_i=a.max_i;
	max_j=a.max_j;
	D_created=a.D_created;
	DPrime_created=a.DPrime_created;
	scoring=a.scoring;
	return a;
}
//=================================================================================================================
//	alignerP()
//=================================================================================================================	
AlignerP::AlignerP(Sequence &seqA_, BasePairs &bpsA_, Sequence &seqB_,
		BasePairs &bpsB_, const AlignerParams *ap_, const Scoring *s_) :
			scoring(s_), params(ap_), seqA(seqA_), bpsA(bpsA_), seqB(seqB_),
			bpsB(bpsB_), r(1, 1, seqA_.length(), seqB_.length()),
			D_created(false), DPrime_created(false),
			alignment(seqA_, seqB_) {

	D.resize(bpsA.num_bps(), bpsB.num_bps());
	D.fill((pf_score_t )0);
	M.resize(seqA.length()+1, seqB.length()+1);
	M.fill((pf_score_t )0);

	DPrime.resize(bpsA.num_bps(), bpsB.num_bps());
	DPrime.fill((pf_score_t )0);

	MPrime.resize(seqA.length()+1, seqB.length()+1);
	MPrime.fill((pf_score_t )0);

	Mrev.resize(seqA.length()+1, seqB.length()+1);
	Mrev.fill((pf_score_t )0);
}

//=================================================================================================================
//	align_noex_inside()
//=================================================================================================================	
//! standard cases in alignment: base match, base in/del, arc match
//! (for structure local alignment this is extended by exclusion handling)
//!  if lonely basepairs are disallowed, there is special treatment
pf_score_t AlignerP::align_noex_inside(size_type al, size_type bl, size_type i, size_type j) {
	//std::cout<<"align_noex called "<<endl;

	//std::cout<<"in align_noex"<<endl;

	pf_score_t max_score;
	//if(i==1 || j==1) { return (pf_score_t)1;} 
	// base match


	max_score = M(i-1, j-1) * scoring->exp_basematch(i, j);
	//if(M(i-1,j-1) <0)
	//    		std::cout<<"in align noex in case 1  :  "<<al<<"  "<<bl<<"  "<<i<<"  "<<j<<endl; 
	// base del

	max_score=max_score+ M(i-1, j) * scoring->exp_gapA(i, j);
	//if(M(i-1,j) <=0)
	//	    		std::cout<<"in align noex in case 2  :  "<<al<<"  "<<al<<"  "<<i<<"  "<<j<<endl; 

	// base ins

	max_score=max_score + M(i, j-1) * scoring->exp_gapB(i, j);
	//if(M(i,j-1) <=0 )
	//	    		std::cout<<"in align noex in case 3 :  "<<al<<"  "<<al<<"  "<<i<<"  "<<j<<endl; 

	// arc match

	// standard case for arc match (without restriction to lonely pairs)
	assert(max_score > 0);
	const BasePairs::RightAdjList &adjlA = bpsA.right_adjlist(i);
	const BasePairs::RightAdjList &adjlB = bpsB.right_adjlist(j);

	// for all pairs of arcs in A and B that have right ends i and j, respectively
	//
	for (BasePairs::RightAdjList::const_iterator arcA=adjlA.begin(); arcA
	!=adjlA.end() && arcA->left() > al; ++arcA) { //was > before
		for (BasePairs::RightAdjList::const_iterator arcB=adjlB.begin(); arcB
		!=adjlB.end() && arcB->left() > bl; ++arcB) {//was > before 

			// consider score for match of basepairs
			assert(M(arcA->left()-1, arcB->left()-1) > 0);
			assert(D(arcA->idx(), arcB->idx())>0);
			//std::cout<<"Printing matrix D in noex inside: "<<endl<<endl;
			//			cout<<D;
			//			std::cout<<endl;
//		std::cout << "accessed D enter in align_noex_in : "<< D(arcA->idx(),arcB->idx())<<endl;
			max_score = max_score + M(arcA->left()-1, arcB->left()-1) * D(arcA->idx(), arcB->idx());

			/* if(max_score<=0) {
			 std::cout<<"in align noex in case 4 :  "<<al<<"  "<<bl<<"  "<<i<<"  "<<j<<endl;
			 std::cout << "M(arcA->l-1,arcB->l-1  "<<M(arcA->left()-1,arcB->left()-1)<<"  D entries " <<arcA->idx()<<"  "<<arcB->idx()<<" " << D(arcA->idx(),arcB->idx())<<endl;
			 exit(-1);
			 } */
		}
	}

	return max_score;
}

//=================================================================================================================
//	init_M_inside()
//=================================================================================================================	
//! generic initalization method. The 0th row and column are initialized
//
void AlignerP::init_M_inside(size_type al, size_type ar, size_type bl, size_type br) {

	M(al, bl)=(pf_score_t)1;

	pf_score_t indel_score=(pf_score_t)1;

	for (size_type i=al+1; i<ar; i++) {
		indel_score *= scoring->exp_gapA(i, bl);
		M(i, bl) = (pf_score_t)indel_score;
		assert(M(i,bl)>0);
	}

	indel_score=(pf_score_t)1;
	for (size_type j=bl+1; j<br; j++) {
		indel_score *= scoring->exp_gapB(al, j);
		M(al, j) = (pf_score_t)indel_score;
	}
}

//=================================================================================================================
//	align_in_arcmatch
//=================================================================================================================	
void AlignerP::align_in_arcmatch(size_type al, size_type ar, size_type bl, size_type br) {

	init_M_inside(al, ar, bl, br);

	for (size_type i=al+1; i<ar; i++)
	  for (size_type j=bl+1; j<br; j++)
	    M(i, j)=align_noex_inside(al,bl,i,j);
}

//=================================================================================================================
//	fill_D_entries_inside
//=================================================================================================================	

// compute the entries in the D matrix that
// can be computed from the matrix/matrices M
// for the subproblem al,bl,max_ar,max_br
//
// pre: M matrices are computed by a call to 
//      align_in_arcmatch(al,max_ar,bl,max_br)
//
void AlignerP::fill_D_entries_inside(size_type al, size_type bl,
		const BasePairs::LeftAdjList &adjlA,
		const BasePairs::LeftAdjList &adjlB, size_type max_ar, size_type max_br) {

	for (BasePairs::LeftAdjList::const_iterator arcA=adjlA.begin(); arcA
	!=adjlA.end(); ++arcA) {

		for (BasePairs::LeftAdjList::const_iterator arcB=adjlB.begin(); arcB
		!=adjlB.end(); ++arcB) {

			size_type ar = arcA->right();
			size_type br = arcB->right();
			//
			// if right ends ar,br exceed the limits max_ar,max_br resp.
			// the corresponding D entry is set to neg_infty
			// in order to dissalow the arc match completely.
			// This occurs only due to the max-diff-am heuristic.
			if (ar>max_ar || br>max_br) {
				D(arcA->idx(), arcB->idx()) = (pf_score_t)0;
			} else {
				pf_score_t m=M(ar-1, br-1); //
				assert(m>0);
				D(arcA->idx(), arcB->idx()) = m * scoring->exp_arcmatch(*arcA,*arcB);

				// std::cout <<"fillD " <<al<<" "<<ar<<";"<<bl<<" "<<br<<": "<<m<<"  "<<D(arcA->idx(),arcB->idx())<< endl;
			}
		}
	}
}

//=================================================================================================================
//	align_D_inside()
//=================================================================================================================	
// compute all entries D
void AlignerP::align_D_inside() {
	// ------------------------------------------------------------
	// General workflow:
	//
	// for all combinations of left arc ends al and ar
	// 1.) determie for which arc-pairs the D entries can be computed   
	// in one run, 2.) call align_in_arcmatch 3.) call fill_D_entries
	// ------------------------------------------------------------

	// ------------------------------------------------------------
	// traverse the left ends al,bl of arcs in descending order
	//
	std::cout<<"Printing matrix D before filling in: "<<endl<<endl;
	cout<<D;
	std::cout<<endl;
	for (size_type al=r.get_endA(); al+1>r.get_startA(); al--) {

		const BasePairs::LeftAdjList &adjlA = bpsA.left_adjlist(al);

		if (adjlA.size()>0) {
			for (size_type bl=r.get_endB(); bl+1>r.get_startB(); bl--) {
				const BasePairs::LeftAdjList &adjlB = bpsB.left_adjlist(bl);
				if (adjlB.size()>0) {
					// ------------------------------------------------------------
					// get maximal right ends of arcs with left ends al,bl 
					// where max_diff_am conditions hold
					// and no_lonely_pairs condition holds
					//
					size_type max_ar=al;
					size_type max_br=bl;
					if (params->max_diff_am<0) {
						max_ar=adjlA[adjlA.size()-1].right();
						max_br=adjlB[adjlB.size()-1].right();
						//        std::cout<<endl<<"in alignD() : maxdiff_am<0";
					} else {
						for (BasePairs::LeftAdjList::const_iterator arcA =
							adjlA.begin(); arcA!=adjlA.end(); ++arcA) {
							for (BasePairs::LeftAdjList::const_iterator arcB =
								adjlB.begin(); arcB!=adjlB.end(); ++arcB) {
								size_type ar = arcA->right();
								size_type br = arcB->right();
								if ((params->max_diff_am<0 || std::max(ar-al, br-bl)-std::min(ar-al, br-bl) <= (size_type)params->max_diff_am)) {
									max_ar=max(max_ar, ar);
									max_br=max(max_br, br);
								}
							}
						}
					}
					// ------------------------------------------------------------
					// align under the maximal pair of arcs
					//
					//   std::cout<<endl<<"calling align_in_arc with : "<<al<<" "<<max_ar<< " : "<<bl<<" "<<max_br<<endl;
					align_in_arcmatch(al, max_ar, bl, max_br);
					// ------------------------------------------------------------
					// fill D matrix entries
					//
					fill_D_entries_inside(al, bl, adjlA, adjlB, max_ar, max_br);

				}

			}
		}

	}
	std::cout<<"Printing matrix D : "<<endl<<endl;
	cout<<D;
	std::cout<<endl;
	D_created=true; // now the matrix D is built up
}

//=================================================================================================================
//	align_inside()
//=================================================================================================================	
//! compute the alignment score

pf_score_t AlignerP::align_inside() {
	// ------------------------------------------------------------
	// computes D matrix and then
	// does the alignment on the top level
	// ------------------------------------------------------------

	if (!D_created) {// std::cout<< "D is going to be created "<<endl; 
		align_D_inside();
	}

	assert(r.get_startA()>0);
	assert(r.get_startB()>0);

	align_in_arcmatch(r.get_startA()-1, r.get_endA()+1, r.get_startB()-1,r.get_endB()+1);
	max_i=r.get_endA();
	max_j=r.get_endB();
	//std::cout<<"in Align() : M(maxi,maxj)  "<<M(max_i,max_j)<<" "<<endl;
	assert(M(max_i, max_j)>0);
	std::cout<<"Printing Matrix M: "<<endl;
	cout<<M;
	std::cout<<endl<<endl;;

	return M(max_i, max_j);

}

//=================================================================================================================
//	align_D_outside()
//=================================================================================================================	
// compute all entries D
void AlignerP::align_D_outside() {
	// ------------------------------------------------------------
	// General workflow:
	//
	// for all combinations of left arc ends al and ar
	// 1.) determine for which arc-pairs the DPrime entries can be computed   
	// in one run, 2.) call align_in_arcmatch 3.) call fill_D_entries
	// ------------------------------------------------------------

	// ------------------------------------------------------------
	// traverse the left ends al,bl of arcs in descending order
	//

	for (size_type al=r.get_startA(); al<=r.get_endA(); al++) {

		const BasePairs::LeftAdjList &adjlA = bpsA.left_adjlist(al);

		if (adjlA.size()>0) {
			for (size_type bl=r.get_startB(); bl<=r.get_endB(); bl++) {
				const BasePairs::LeftAdjList &adjlB = bpsB.left_adjlist(bl);
				if (adjlB.size()>0) {
					// ------------------------------------------------------------
					// get minimal right ends of arcs with left ends al,bl 
					// where max_diff_am conditions hold
					// and no_lonely_pairs condition holds
					//
				        size_type min_ar=r.get_endA();
					size_type min_br=r.get_endB();
					
					assert(params->max_diff_am<0);
					if (params->max_diff_am<0) {
					  min_ar=adjlA[0].right();
					  min_br=adjlB[0].right();
					  //    std::cout<<endl<<"in alignD() : maxdiff_am<0";
					} else {
						for (BasePairs::LeftAdjList::const_iterator arcA =
							adjlA.begin(); arcA!=adjlA.end(); ++arcA) {
							for (BasePairs::LeftAdjList::const_iterator arcB =
								adjlB.begin(); arcB!=adjlB.end(); ++arcB) {

								size_type ar = arcA->right();
								size_type br = arcB->right();
								std::cout<<" ar and br "<<ar<<"  "<<br<<endl;
								if (std::max(ar-al, br-bl)-std::min(ar-al, br-bl) <= (size_type)params->max_diff_am) {
									min_ar=min(min_ar, ar);
									min_br=min(min_br, br);

							      }
							}
						}
					}
					
					// for debugging only
					MPrime.fill((pf_score_t )0.0);

					// ------------------------------------------------------------
					// align outside the arc
					align_out_arcmatch(al, min_ar, bl, min_br);
					std::cout<<"Printing Matrix MPrime: IN align_D_out "<<al<<" "<<min_ar<<"  "<<bl<<" "<<min_br<<endl;
	
					cout<<MPrime;
					std::cout<<endl<<endl;;

					// ------------------------------------------------------------
					// fill DPrime matrix entries
					//
					fill_D_entries_outside(al, bl, adjlA, adjlB, min_ar, min_br);
					std::cout<<"Printing Matrix DPrime: IN align_D_out "<<al<<"  "<<bl<<endl;
	
					cout<<DPrime;
					std::cout<<endl<<endl;;


				}

			}
		}

	}
	DPrime_created=true; // now the matrix DPrime is built up
}

//=================================================================================================================
//	align_noex_outside()
//=================================================================================================================	
//! standard cases in alignment: base match, base in/del, arc match
//! compute partition function for alignments with hole al..i and bl..j (recursing to larger holes)
pf_score_t AlignerP::align_noex_outside(size_type al, size_type bl,
		size_type i, size_type j) {

	/* if(i>=r.get_endA() && j>=r.get_endB()){
	 //std::cout<<"in noex_outside with al,bl;ar,br"<<al<<" "<<bl<<" ; "<<i<<"  "<<j<<endl;
	 return (pf_score_t)1;}*/

	pf_score_t max_score;

	// base match
	assert(MPrime(i+1, j+1) > 0);
	max_score = MPrime(i+1, j+1) * scoring->exp_basematch(i+1, j+1);
	//if(max_score <0)
	//std::cout<<"Max score of outside up to case 1: " <<"  "<<al<<"  "<<bl<<"  "<<i<<"  "<<j<<" "<<scoring->exp_basematch(i+1,j+1)<<endl;
	//if(max_score > 0)
	//std::cout<<"Max score of outside up to case 1: " << MPrime(i+1,j+1)<<"   " << scoring->exp_basematch(i+1,j+1)<<endl;
	// base del
	assert(MPrime(i+1, j) > 0);
	max_score=max_score+ MPrime(i+1, j) * scoring->exp_gapA(i+1, j);
	// if(max_score <0)
	//std::cout<<"Max score of outside up to case 2: " <<"  "<<al<<"  "<<bl<<"  "<<i<<"  "<<j<<" "<<MPrime(i+1,j)<<endl;
	// base ins
	assert(MPrime(i, j+1) > 0);
	max_score=max_score + MPrime(i, j+1) * scoring->exp_gapB(i, j+1);
	//std::cout<<"Max score of outside up to case 3: " <<max_score<<endl;
	// arc match-case 4
	// standard case for arc match (without restriction to lonely pairs)
	//if(max_score <0)
	// std::cout<<"Max score of outside up to case 3: " <<"  "<<al<<"  "<<bl<<"  "<<i<<"  "<<j<<"  "<<MPrime(i,j+1)<<endl;

	{
		const BasePairs::RightAdjList &adjlA = bpsA.right_adjlist(i+1);
		const BasePairs::RightAdjList &adjlB = bpsB.right_adjlist(j+1);

		// for all pairs of arcs in A and B that have right ends i+1 and j+1, respectively
		//
		//std::cout<<"Rightadj list of "<<i+1<<endl;
		//for(BasePairs::RightAdjList::const_iterator arcA=adjlA.begin(); arcA!=adjlA.end();++arcA){
		//  if (arcA->left() > al)
		// std::cout<<"Left end index "<<arcA->left()<<"and al"<<al<<endl;  }

		for (BasePairs::RightAdjList::const_reverse_iterator arcA=adjlA.rbegin();
		arcA!=adjlA.rend() && arcA->left() < al; ++arcA) {
			for (BasePairs::RightAdjList::const_reverse_iterator arcB=adjlB.rbegin();
			arcB!=adjlB.rend() && arcB->left() < bl; ++arcB) {
				// consider score for match of basepair
				//std::cout<<"Mrev(arcA->left()+1,arcB->left()+1) "<<Mrev(arcA->left()+1,arcB->left()+1)<<endl;
				
			        assert(DPrime(arcA->idx(),arcB->idx()) > 0);
				assert(Mrev(arcA->left()+1,arcB->left()+1) > 0);
							
				max_score += 
				  DPrime(arcA->idx(),arcB->idx())
				  * scoring->exp_arcmatch(*arcA,*arcB) 
				  * Mrev(arcA->left()+1,arcB->left()+1);
			}
		}
	}

	//std::cout<<"Max score of outside up to case 4: " << max_score <<"  "<<al<<"  "<<bl<<"  "<<i<<"  "<<j<<endl;


	{
		const BasePairs::LeftAdjList &adjlA = bpsA.left_adjlist(i+1);
		const BasePairs::LeftAdjList &adjlB = bpsB.left_adjlist(j+1);

		// arc match-case 5
		// standard case for arc match (without restriction to lonely pairs)
		// for all pairs of arcs in A and B that have left ends i and j, respectively

		for (BasePairs::LeftAdjList::const_iterator arcA=adjlA.begin(); arcA!=adjlA.end(); ++arcA) {
			for (BasePairs::LeftAdjList::const_iterator arcB=adjlB.begin(); arcB!=adjlB.end(); ++arcB) {
				// consider score for match of basepairs
				//std::cout << max_score << " " << MPrime(arcA->right(),arcB->right()) * D(arcA->idx(),arcB->idx())<< std::endl;

				assert(MPrime(arcA->right(),arcB->right())>0);
				assert(D(arcA->idx(),arcB->idx())>0);

				max_score += MPrime(arcA->right(),arcB->right()) * D(arcA->idx(),arcB->idx());
				//std::cout << "R" << max_score << std::endl;
			}
		}
		//std::cout<<"Max score of outside up to case 5: " <<max_score<<endl;
	}

	//std::cout<<"Max score of outside up to case 5: " <<"  "<<al<<"  "<<bl<<"  "<<i<<"  "<<j<<endl;
	return max_score;
}
//=================================================================================================================
//	init_M_outside()
//=================================================================================================================	
//! generic initalization method. The 0th row and column are initialized
//
void AlignerP::init_M_outside(size_type al, size_type ar, size_type bl, size_type br) {

	assert(M(al,bl)>0);

	MPrime(r.get_endA(),r.get_endB())=M(al,bl);
	//std::cout<<"in init_M_out :  with al,ar,bl,br "<<al<<" "<<ar<< " : "<<bl<<" "<<br<<endl;

	for (int i=(int)r.get_endA()-1;i>=(int)ar; i--) {
	  assert(MPrime(i+1,r.get_endB())>0);
	  MPrime(i,r.get_endB()) = MPrime(i+1,r.get_endB()) * scoring->exp_gapA(i+1,r.get_endB());
	}
	
	for (int j=(int)r.get_endB()-1; j>=(int)br; j--) {
	  assert(MPrime(r.get_endA(),j+1)>0);
	  MPrime(r.get_endA(),j) = MPrime(r.get_endA(),j+1) * scoring->exp_gapB(r.get_endA(),j+1);;
	}


	//std::cout<<"in init_M_out "<< MPrime(r.get_endA(),r.get_endB())<<"for al, bl "<<al<<"  "<<bl<<" ar,br"<<ar<<" "<<br<<endl;

	Mrev(al,bl)=(pf_score_t)1;
	pf_score_t indel_score=(pf_score_t)1;
	
	if (al>0) {
	  for (size_type i=al-1; i>=r.get_startA(); i--) {
	    assert(Mrev(i+1,bl)>0);
	    Mrev(i,bl) = M(i+1,bl) * scoring->exp_gapA(i,bl);;
	  }
	}
	if (bl>0) {
	  indel_score=(pf_score_t)1;
	  for (size_type j=bl-1; j>=r.get_startB(); j--) {
	    assert(Mrev(al,j+1)>0);
	    Mrev(al,j)= Mrev(al,j+1) * scoring->exp_gapB(al,j);
	  }
	}
}

//=================================================================================================================
//	align_out_arcmatch
//=================================================================================================================	

//! fill MPrime(i,j) entries for fixed left indices al and bl
//! where i>=ar and j>=br
void
AlignerP::align_out_arcmatch(size_type al,size_type ar,size_type bl,size_type br) {
	//std::cout<<"in align out arcmatch : calling init_out "<<endl;

	init_M_outside(al,ar,bl,br);

	if(al>=r.get_startA() && bl>=r.get_startB()) { // don't compute reversed M for al=startA-1 or bl=startB-1 

		for(size_type i=al-1; i>=r.get_startA();i--) {
			for(size_type j=bl-1; j>=r.get_startB();j--) {
				Mrev(i,j)=align_noex_reversedM(i,j,al,bl);
			}
		}
	}

	for(int i=(int)r.get_endA()-1; i>=(int)ar; i--) {
	  for(int j=(int)r.get_endB()-1;j>=(int)br; j--) {
			MPrime(i,j) = align_noex_outside(al,bl,i,j);
		}
	}

}

//=================================================================================================================
//	fill_D_entries_outside
//=================================================================================================================	

// compute the entries in the DPrime matrix that
// can be computed from the matrix/matrices MPrime
// for the subproblem al,bl,min_ar,min_br
//
// pre: MPrime matrices are computed by a call to 
//      align_out_arcmatch(al,min_ar,bl,min_br)
//
void
AlignerP::fill_D_entries_outside(size_type al, size_type bl,
		const BasePairs::LeftAdjList &adjlA,
		const BasePairs::LeftAdjList &adjlB,
		size_type min_ar, size_type min_br)
{

	for (BasePairs::LeftAdjList::const_iterator arcA=adjlA.begin();
	arcA!=adjlA.end(); ++arcA) {

		for (BasePairs::LeftAdjList::const_iterator arcB=adjlB.begin();
		arcB!=adjlB.end(); ++arcB) {

			size_type ar = arcA->right();
			size_type br = arcB->right();
			//
			// if right ends ar,br exceed the limits min_ar,min_br resp.
			// the corresponding DPrime entry is set to zero
			// in order to dissalow the arc match completely.
			// This occurs only due to the max-diff-am heuristic.
			if (ar<min_ar || br<min_br ) {
				DPrime(arcA->idx(),arcB->idx()) = (pf_score_t)0;
			} else {
				assert(MPrime(ar,br)>0);
				pf_score_t m=MPrime(ar,br); 
				DPrime(arcA->idx(),arcB->idx()) = m;
				//std::cout <<al<<" "<<ar<<";"<<bl<<" "<<br<<": "<<MPrime(ar-1,br-1)<<" "<<scoring->arcmatch(*arcA,*arcB)<<" "<<DPrime(arcA->idx(),arcB->idx())<< endl;
			}
			assert(DPrime(arcA->idx(),arcB->idx())>0);
		}
	}
}

//=================================================================================================================
//	align_outside()
//=================================================================================================================	
//! compute the alignment score
pf_score_t AlignerP::align_outside() {
	// ------------------------------------------------------------
	// computes DPrime matrix and then
	// does the alignment on the top level
	// ------------------------------------------------------------

	if (!DPrime_created) { //std::cout<< "DPrime is going to be created "<<endl; 
		align_D_outside();}

	align_out_arcmatch(r.get_startA()-1,r.get_startA()-1,r.get_startB()-1,r.get_startB()-1);
	//	std::cout<<"in Align out() : starts "<<r.get_startA()<<"  "<<r.get_startB()<<endl; 
	
	std::cout<<"Printing Matrix MPrime: "<<endl;
	
	cout<<MPrime;
	std::cout<<endl<<endl;;
	return MPrime(r.get_startA()-1,r.get_startB()-1); 
}

//=================================================================================================================
//	align_noex_reversedM()
//=================================================================================================================	
//! standard cases in alignment for reversed case: base match, base in/del, arc match
//! compute pf of alignments i..al-1 and j..bl-1 
pf_score_t
AlignerP::align_noex_reversedM(size_type i, size_type j,size_type al, size_type bl) {

	pf_score_t max_score;
	// base match
	assert(Mrev(i+1,j+1)>0);
	max_score = Mrev(i+1,j+1) * scoring->exp_basematch(i,j);

	//std::cout<<"max score in reversed M 1st so far"<<max_score<<"for i j : "<<i<<" "<<j<<endl;
	// base del
	assert(Mrev(i+1,j)>0);
	max_score=max_score+ Mrev(i+1,j) * scoring->exp_gapA(i,j-1);
	//std::cout<<"max score in reversed M 2nd  so far"<<max_score<<"for i j : "<<i<<" "<<j<<endl;
	// base ins
	assert(Mrev(i,j+1)>0);
	max_score=max_score + Mrev(i,j+1) * scoring->exp_gapB(i-1,j);
	//std::cout<<"max score in reversed M so far"<<max_score<<endl;
	// arc match
	// standard case for arc match (without restriction to lonely pairs)
	const BasePairs::LeftAdjList &adjlA = bpsA.left_adjlist(i);
	const BasePairs::LeftAdjList &adjlB = bpsB.left_adjlist(j);

	// for all pairs of arcs in A and B that have right ends i and j, respectively
	//
	for (BasePairs::LeftAdjList::const_iterator arcA=adjlA.begin();
	arcA!=adjlA.end() && arcA->right() < al; ++arcA) {
		for (BasePairs::LeftAdjList::const_iterator arcB=adjlB.begin();
		arcB!=adjlB.end() && arcB->right() < bl; ++arcB) {

			// consider score for match of basepairs
			assert(Mrev(arcA->right()+1,arcB->right()+1)>0);
			assert(D(arcA->idx(),arcB->idx())>0);
			max_score =
				Mrev(arcA->right()+1,arcB->right()+1) * D(arcA->idx(),arcB->idx());
			//std::cout<<"Mrev in align_no_exrev: THEN "<<max_score<<"  "<<Mrev(arcA->right()+1,arcB->right()+1)<<"  "<<D(arcA->idx(),arcB->idx())<<endl;		   	
		}
	}

	return max_score;
}

