
/**********************************************************************
 *
 * LocARNA: LOCal Alignment of RNA
 *
 * Copyright (C) Sebastian Will <will(@@)informatik.uni-freiburg.de> 
 *               2005, 2006, 2007
 *
 **********************************************************************/


#include <iostream>
#include <fstream>
// #include <sstream>
#include <vector>

#include <math.h>

#include "sequence.hh"
#include "basepairs.hh"
#include "alignment.hh"
#include "aligner_p.hh"

#include "rna_data.hh"

#include "match_probs.hh"
#include "ribosum.hh"

#include "ribosum85_60.icc"

#include "anchor_constraints.hh"

using namespace std;

const std::string 
VERSION_STRING = (std::string)PACKAGE_STRING; 

// ------------------------------------------------------------
// Parameter

double min_prob; // only pairs with a probability of at least min_prob are taken into account

int match_score;
int mismatch_score;
int indel_score;
int indel_opening_score;
int temperature;
int struct_weight;

int tau_factor; // contribution of sequence similarity in an arc match (in percent)
               

bool struct_local; // allow exclusions for maximizing alignment of connected substructures 
bool sequ_local; // maximize alignment of subsequences

const bool DO_TRACE=true;


int max_diff; // maximal difference for positions of alignment edges
              // (only used for ends of arcs)
int max_diff_am; //maximal difference between two arc ends, -1 is off

// only consider arc matchs where
//   1. for global (bl-al)>max_diff || (br-ar)<=max_diff    (if max_diff>=0)
//   2. for local (ar-al)-(br-bl)<=max_diff_am              (if max_diff_am>=0)
// except when there is no additional computation of M matrices necessary,
// this occurs if arcs are left-incident with larger arcs where 1 and 2 hold


int exclusion_score; // Score contribution per exclusion
                     // set to zero for unrestricted structure locality

double prob_exp_f(int seqlen) {return 1.0/(2*seqlen);} // expected probability of a base pair (null-model)

double exp_prob;
bool opt_exp_prob;

int output_width=70;

// ------------------------------------------------------------
// File arguments
std::string bpsfile1;
std::string bpsfile2;

std::string clustal_out;
bool opt_clustal_out;

std::string pp_out;
bool opt_pp_out;

// ------------------------------------------------------------
//
// Options
//
#include "options.hh"


bool opt_help;
bool opt_version;
bool opt_verbose;
bool opt_local_output;
bool opt_pos_output;

std::string ribosum_file;
bool use_ribosum;

option_def my_options[] = {
    {"help",'h',&opt_help,O_NO_ARG,0,O_NODEFAULT,"","This help"},
    {"version",'V',&opt_version,O_NO_ARG,0,O_NODEFAULT,"","Version info"},
    {"verbose",'v',&opt_verbose,O_NO_ARG,0,O_NODEFAULT,"","Verbose"},
    
    {"",0,0,O_SECTION,0,O_NODEFAULT,"","Scoring parameters"},
    
    {"match",'m',0,O_ARG_INT,&match_score,"50","score","Match score"},
    {"mismatch",'M',0,O_ARG_INT,&mismatch_score,"0","score","Mismatch score"},
    {"ribosum-file",0,0,O_ARG_STRING,&ribosum_file,"RIBOSUM85_60","f","Ribosum file"},
    {"use-ribosum",0,0,O_ARG_BOOL,&use_ribosum,"true","bool","Use ribosum scores"},
    {"indel",'i',0,O_ARG_INT,&indel_score,"-350","score","Indel score"},
    {"indel-opening",0,0,O_ARG_INT,&indel_opening_score,"-500","score","Indel opening score"},
    {"struct-weight",'s',0,O_ARG_INT,&struct_weight,"180","score","Maximal weight of 1/2 arc match"},
    {"exp-prob",'e',&opt_exp_prob,O_ARG_DOUBLE,&exp_prob,O_NODEFAULT,"prob","Expected probability"},
    {"tau",'t',0,O_ARG_INT,&tau_factor,"0","factor","Tau factor in percent"},
    {"temperature",0,0,O_ARG_INT,&temperature,"800","int","Temperature for PF-computation"},
    {"",0,0,O_SECTION,0,O_NODEFAULT,"","Type of locality"},
 
    {"min-prob",'p',0,O_ARG_DOUBLE,&min_prob,"0.0005","prob","Minimal probability"},
    {"max-diff-am",'D',0,O_ARG_INT,&max_diff_am,"-1","diff","Maximal difference for sizes of matched arcs"},
          
   // {"kbest",'k',0,O_ARG_INT,&kbest_k,"-1","k","Find k-best alignments"},
 
    {"",0,0,O_SECTION,0,O_NODEFAULT,"","Controlling output"},
 
    {"width",'w',0,O_ARG_INT,&output_width,"120","columns","Output width"},
    
    {"",0,0,O_SECTION,0,O_NODEFAULT,"","Heuristics for speed accuracy trade off"},
    
    {"max-diff-match",'d',0,O_ARG_INT,&max_diff,"-1","diff","Maximal difference for alignment edges"},
    
    {"",0,0,O_SECTION,0,O_NODEFAULT,"","RNA sequences and pair probabilities"},
    
    {"",0,0,O_ARG_STRING,&bpsfile1,O_NODEFAULT,"bps-file 1","Basepairs input file 1"},
    {"",0,0,O_ARG_STRING,&bpsfile2,O_NODEFAULT,"bps-file 2","Basepairs input file 2"},
    {"",0,0,0,0,O_NODEFAULT,"",""}
};


// ------------------------------------------------------------


// ------------------------------------------------------------
// MAIN

int
main(int argc, char **argv) {
    
    // ------------------------------------------------------------
    // Process options
    
    bool process_success=process_options(argc,argv,my_options);
    std::cout<<"in main options";
    if (opt_help) {
	cout << VERSION_STRING<<endl;
	
	cout << "Copyright Sebastian Will, 2004-2008"<<endl<<endl;
	
	cout << "A tool for pairwise Local (and global) Alignment of RNA."<<endl<<endl;
	
	print_help(argv[0],my_options);
	
	cout << "Report bugs to <will (at) informatik.uni-freiburg.de>."<<endl<<endl;
	exit(0);
    }
    
    if (opt_version || opt_verbose) {
	cout << VERSION_STRING<<endl;
	if (opt_version) exit(0); else cout <<endl;
    }

    if (!process_success) {
	puts(O_error_msg);
	printf("USAGE: ");
	print_usage(argv[0],my_options);
	printf("\n");
	exit(-1);
    }
        
    if (opt_verbose)
	print_options(my_options);
    

    // ------------------------------------------------------------
    // Get input data and generate data objects
    //
    std::cout<<"in main ";
    BasePairsParams bp_params(min_prob);

    RnaData rnadataA(bpsfile1);
    
    RnaData rnadataB(bpsfile2);
    std::cout<<"Objects instantiated";
    Sequence seqA=rnadataA.get_sequence();
    Sequence seqB=rnadataB.get_sequence();
    
    BasePairs bpsA(&rnadataA,&bp_params);
    BasePairs bpsB(&rnadataB,&bp_params);
    // ----------------------------------------
        // report on input in verbose mode
        if (opt_verbose) {
    	cout << "Sequence A: "<<endl;
    	seqA.write(cout);
    	cout<<" (Length:"<< seqA.length()<<", Basepairs:"<<bpsA.num_bps() << ")" <<endl;
    		
    	cout << "Sequence B: "<<endl;
    	seqB.write(cout);
    	cout<<" (Length:"<< seqB.length()<<", Basepairs:"<<bpsB.num_bps() << ")" <<endl;

    	// cout << "Base Identity: "<<(seq_identity(seqA,seqB)*100)<<endl; 
        }
        
        // ----------------------------------------  
           // Ribosum matrix
           //
           Ribosum *ribosum=NULL;
           
           if (use_ribosum) {
       	if (ribosum_file == "RIBOSUM85_60") {
       	    if (opt_verbose) {
       		std::cout <<"Use built-in ribosum."<<std::endl;
       	    }
       	    ribosum = new Ribosum85_60();
       	} else {
       	    ribosum = new RibosumFreq(ribosum_file);
       	}
           }
       
      //  MatchProbs *match_probs=0L;
        
   // ----------------------------------------
  // construct scoring
        if (opt_verbose) {
      	std::cout << "Construct scoring."<<std::endl;
          }
          
          ScoringParams scoring_params(match_score,
				       mismatch_score,
				       indel_score,
				       indel_opening_score,
				       ribosum,
				       struct_weight,
				       tau_factor,
				       0, // exclusion score
				       opt_exp_prob?exp_prob:-1,
				       temperature,//temperature,
				       false,//opt_stacking,
				       false,//opt_mea_alignment,
				       0,//match_probs,
				       0,//mea_alpha,
				       0,//mea_beta,
				       0//mea_gamma,
				        );
          

          		  
          	Scoring scoring(&bpsA,&bpsB,&scoring_params);
          	AnchorConstraints seq_constraints(seqA.length(),"",seqB.length(),"");
    

          // ------------------------------------------------------------
             // Computation of the alignment score
            //
           // parameter for the alignment
	     AlignerParams aligner_params(false,//no_lonely_pairs,
				       false,//struct_local,
				       false,//sequ_local,
				       max_diff,
				       max_diff_am,
				       false,//opt_stacking,
				       seq_constraints);
             
             // initialize aligner object, which does the alignment computation
         AlignerP aligner(seqA,bpsA,seqB,bpsB,&aligner_params,&scoring);
        pf_score_t score = aligner.align_inside();
         
             // ----------------------------------------
             // report score
             //
            std::cout << "Score through inside algorithm: "<<score<<std::endl;
	    score = aligner.align_outside();
	    std::cout << "Score through outside algorithm: "<<score<<std::endl;
 
    // DONE
    exit(0);
}

